// $Id: DAMLModelImpl.java,v 1.17 2003/03/11 22:55:58 drager Exp $

// This code is based on NodeCenteredView by Mike Dean.

// DAMLView builds upon:
// - The RDF-API statement centered Model.
// - Mike Dean's NodeCenteredView
// and provides some built in semantics from DAML - equivalence, class/sub-class relations, others?

package org.daml.validator.model;

//import org.w3c.rdf.util.*;
//import org.w3c.rdf.model.*;

import com.hp.hpl.mesa.rdf.jena.model.Resource;
import com.hp.hpl.mesa.rdf.jena.model.Property;
import com.hp.hpl.mesa.rdf.jena.model.RDFNode;
import com.hp.hpl.mesa.rdf.jena.model.Statement;
import com.hp.hpl.mesa.rdf.jena.model.Model;
import com.hp.hpl.mesa.rdf.jena.model.Literal;
import com.hp.hpl.mesa.rdf.jena.model.StmtIterator;
import com.hp.hpl.mesa.rdf.jena.model.RDFException;

import com.hp.hpl.mesa.rdf.jena.mem.ModelMem;

import java.util.*;
import org.daml.validator.Preferences;
import org.daml.validator.DAMLDatatypeValidator;
import org.daml.validator.xsd.UnknownSchemaException;
import org.daml.validator.xsd.SchemaValidationException;
import org.daml.validator.xsd.UnknownSimpleTypeException;

/** 
 * DAML model used by the validator.  Supports a node based view of the RDF-API model
 * with built in semantics from RDF and DAML.  Contains code from NodeCenteredView by 
 * Mike Dean.
 * @author David Rager 
 * @author drager@bbn.com
 */

public class DAMLModelImpl
{
    /**
     * A RDF factory for creating new Models.
     */
    //static org.w3c.rdf.util.RDFFactory factory = new org.w3c.rdf.util.RDFFactoryImpl();

    /**
     * The Model this DAMLModel is based on.
     */
    public Model model;

    /**
     * The XML Schema Datatype Validator
     */
    DAMLDatatypeValidator datatypeValidator;

    /**
     * The preference file
     */
    Preferences preferences;

    /**
     * Collection of predicates (RDF Resources) that have an equivalence semantic.
     */
    public Vector equivalentPredicates = new Vector();

    /**
     * Collection of predicates (RDF Resources) that have a subClass semantic.
     */
    public Vector subClassOfPredicates = new Vector();

    /**
     * Collection of predicates (RDF Resources) that represent domain
     */
    public Vector domainPredicates = new Vector();

    /**
     * Collection of predicates (RDF Resources) that represent range
     */
    public Vector rangePredicates = new Vector();

    /**
     * Collection of predicates (RDF Resources) that represent the imports predicate
     */
    public Vector importsPredicates = new Vector();

    /**
     * Collection of classes (RDF Resources) that represent the Restriction class
     */
    public Vector restrictionClasses = new Vector();

    public Vector objectRestrictionClasses = new Vector();
    
    public Vector datatypeRestrictionClasses = new Vector();

    public Vector classClasses = new Vector();

    public Vector propertyClasses = new Vector();
    
    public Vector datatypePropertyClasses = new Vector();
    
    public Vector objectPropertyClasses = new Vector();

    public Vector allPropertyClasses = new Vector();
    
    public Vector cardinalityPredicates = new Vector();

    public Vector cardinalityQPredicates = new Vector();
    
    public Vector hasClassPredicates = new Vector();

    public Vector hasClassQPredicates = new Vector();

    public Vector hasValuePredicates = new Vector();
    
    public Vector maxCardinalityPredicates = new Vector();
    
    public Vector maxCardinalityQPredicates = new Vector();
    
    public Vector minCardinalityPredicates = new Vector();
    
    public Vector minCardinalityQPredicates = new Vector();
    
    public Vector onPropertyPredicates = new Vector();
    
    public Vector toClassPredicates = new Vector();

    public Vector typePredicates = new Vector();

    public Vector oneOfPredicates = new Vector();

    public Vector complementOfPredicates = new Vector();
    
    public Vector unionOfPredicates = new Vector();

    public Vector intersectionOfPredicates = new Vector();

    public Vector valuePredicates = new Vector();

    public Vector subPropertyOfPredicates = new Vector();

    public Vector ancestorPredicates = new Vector();

    public Vector samePropertyAsPredicates = new Vector();

    public Vector equivalentToPredicates = new Vector();

    public Vector literalClasses = new Vector();
    
    public Vector datatypeClasses = new Vector();

    public Vector resourceClasses = new Vector();

    public Vector listClasses = new Vector();

    public Class DAMLVocabulary;

    //private String DAML_VOCABULARY = "org.daml.rdf.vocabulary.daml_oil_200110.DAML_OIL";
    private String DAML_VOCABULARY = "org.daml.rdf.vocabulary.daml_oil_20010327.DAML_OIL";
    //private String RDFS_VOCABULARY = "org.w3c.rdf.vocabulary.rdf_schema_20001.RDFS";

    /**
     * A Node exists for a RDF-API Resource and contains all incoming
     * and outgoing predicates as well as any equivalent Nodes.
     */

    public class Node
    {
	/**
	 * Resource from RDF-API model.
	 */
	public Resource resource;

	/**
	 * All Statements where this node is the object.
	 */
	public Model incoming = new ModelMem();

	/**
	 * All Statements where this node is the subject.
	 */
	public Model outgoing = new ModelMem();

	/**
	 * All Nodes that are equivalent to this node.
	 */
	Set equivalent = new HashSet();

	/**
	 * Immediate super-classes.
	 */
	Set superclasses = new HashSet();

	/**
	 * Immediate sub-classes.
	 */
	Set subclasses = new HashSet();

	/**
	 * Nodes are created from a RDF-API Resource.
	 */
	Node(Resource resource) 
	{ 
	    this.resource = resource;
	}


	/**
	 * returns a single Value (a NodeCenteredView.Node or a String)
	 * based on the specified path expression.  If no such path is 
	 * found, null is returned.
	 */
	Object findAny(String path)
	    throws Exception
	{
	    Object retval = null;
	    Statement statement;

	    // parse path
	    int end;
	    for (end = 1; end < path.length(); end++)
		{
		    if ((path.charAt(end) == '<')
			|| (path.charAt(end) == '>'))
			break;
		}
	    String predicate = path.substring(1, end);

	    if (path.charAt(0) == '<')
		{
		    statement = findAnyStatement(incoming,
						 predicate);
		    if (statement != null)
			retval = statement.getSubject();
		}
	    else if (path.charAt(0) == '>')
		{
		    statement = findAnyStatement(outgoing,
						 predicate);
		    if (statement != null)
			retval = statement.getObject();
		}

	    // convert
	    if (retval instanceof Resource)
		retval = getNode((Resource) retval);
	    else if (retval instanceof Literal)
		retval = retval.toString();

	    // process any continuation
	    if ((retval != null) 
		&& (end < path.length()))
		{
		    retval = ((Node) retval).findAny(path.substring(end));
		}
	    
	    return retval;
	}


	/**
	 * Returns all statements that match the given query.
	 */
	protected void findAllInternal(String path,
				     java.util.Vector retval)
	    throws RDFException
	{
	    // parse path
	    int end;
	    for (end = 1; end < path.length(); end++)
		{
		    if ((path.charAt(end) == '<')
			|| (path.charAt(end) == '>'))
			break;
		}
	    String predicate = path.substring(1, end);
	    String match = "#" + predicate;

	    Model model = (path.charAt(0) == '<') ? incoming : outgoing;
	    for (StmtIterator en = model.listStatements(); en.hasNext();)
		{
		    Statement statement = (Statement) en.next();
		    String thisPredicate = statement.getPredicate().toString();
		    if (thisPredicate.endsWith(match))
			{
			    Object value = (path.charAt(0) == '<') ? statement.getSubject() : statement.getObject();
			    
			    // convert
			    if (value instanceof Resource)
				value = getNode((Resource) value);
			    else if (value instanceof Literal)
				value = value.toString();

			    // process any continuation
			    if (end < path.length())
				{
				    ((Node) value).findAllInternal(path.substring(end),
								   retval);
				}
			    else
				{
				    retval.add(value);
				}
			}
		}
	}

	/**
	 * returns a (possibly empty) Vector of
	 * NodeCenteredView.Node or String
	 * based on the specified path expression.
	 */
	java.util.Vector findAll(String path)
	    throws Exception
	{
	    java.util.Vector retval = new java.util.Vector();
	    findAllInternal(path,
			    retval);
	    return retval;
	}


	/**
	 * Returns the string representation of the contained Resource.
	 */
	public String toString()
	{
	    return resource.toString();
	}

	/**
	 * Prints out information about this node.
	 */
	public void printInfo()
	{
	    System.out.println("[Node: "+this);
	    System.out.println("equivalent: "+this.equivalent);
	    System.out.println("superclasses: "+this.superclasses);
	    System.out.println("subclasses: "+this.subclasses);
	    System.out.println("outgoing: ");
	    try {
		for (StmtIterator i = outgoing.listStatements(); i.hasNext(); System.out.println(i.next()));
	    } catch (RDFException me) {
		System.out.println(me);
	    }
	    System.out.println("incoming: ");
	    try {
		for (StmtIterator i = incoming.listStatements(); i.hasNext(); System.out.println(i.next()));
	    } catch (RDFException me) {
		System.out.println(me);
	    }
	    System.out.println("]");
	}

	/**
	 * Returns true if the nodes are equal (refer to the same URI?).
	 */
	public boolean equals(Node node)
	{	
	    return (this.resource.equals(node.resource));
	}

	public boolean equals(Resource resource)
	{
	    return (this.resource.equals(resource));
	}

	public boolean equals(RDFNode rdfnode)
	{
	    return (this.resource.equals(rdfnode));
	}

	/**
	 * Returns true if the nodes are equal or declared equivalent.
	 * Assumes that all equivalent nodes are in the equivalence list.
	 */
	public boolean equivalentTo(Node node)
	{
	    // if the nodes are equal then they are equivalent
	    if (this.equals(node)) {
		return(true);
	    } else {
		// if the node equals a node in the equivalence list then they are equivalent
		for (Iterator en = equivalent.iterator(); en.hasNext(); ) {
		    if (node.equals((Node)en.next())) {
			return(true);
		    }
		}
		// otherwise they are not equivalent
		return(false);
	    }
	}

	/**
	 * Returns the type of a Node if one exists.
	 */
	// NOTE: This should be rewritten to do a more accurate job at determining the type
	// NOTE: What about Literals?
	// If more than one type, then it is most likely an instance Node.  RDFS allows resources
	// to be an instance of more than one class.  Class nodes should only be of one type.
	public Node getType()
	{
	    Object type = null;
	    try {
		type = this.findAny(">type");
	    } catch (Exception ex) {
		// ignore error
	    }

	    if (type == null) {
		return(null);
	    } else if (type instanceof DAMLModelImpl.Node) {
		return((DAMLModelImpl.Node)type);
	    } else {
		return(null);
	    }
	}

	/**
	 * Returns a vector of Nodes that are the type classes of this Node.
	 */
	public Set getTypes()
	{
	    HashSet result = new HashSet();
	    Vector stmts = getAllStatements(this.outgoing, typePredicates);
	    //Vector stmts = getOutgoingStatements(typePredicates, true, true);
	    for (Enumeration e = stmts.elements(); e.hasMoreElements(); ) {
		Statement statement = (Statement)e.nextElement();
		Object object = statement.getObject();
		if (object instanceof Resource) {
		    result.add(getNode((Resource)object));
		}
	    }
	    return(result);
	}

	/**
	 * Returns a set of statements that assert the type classes of this Node.
	 */
	public Set getTypeStatements()
	{
	    HashSet result = new HashSet();
	    //Vector stmts = getAllStatements(this.outgoing, typePredicates);
	    Vector stmts = getOutgoingStatements(typePredicates, true, true);
	    for (Enumeration e = stmts.elements(); e.hasMoreElements(); ) {
		Statement statement = (Statement)e.nextElement();
		Object object = statement.getObject();
		if (object instanceof Resource) {
		    result.add(statement);
		}
	    }
	    return(result);
	}
	    

	/**
	 * Returns true if the node is a subclass of the given node.
	 * Takes into account equivalence between nodes.
	 */
	public boolean subClassOf(Node superclass)
	{
	    Vector v = new Vector();
	    v.add(superclass);
	    return(subClassOf(v));
	}
	    
	/**
	 * Returns true if the node is a subclass of the given Resource.
	 * Takes into account equivalence between nodes.
	 */
	public boolean subClassOf(Resource superclass)
	{
	    Vector v = new Vector();
	    Node node = getNode(superclass, false);
	    if (node == null) {
		return false;
	    } else {
		v.add(node);
		return(subClassOf(v));
	    }
	}

	/**
	 * Returns true if the node is a subclass of the given collection of Resources.
	 * Takes into account equivalence between nodes.
	 */
	public boolean subClassOf(Collection superclass)
	{

	    /* Algorithm: Starting with "this" node and work up the
	     * isa hierarchy, and across the equivalence links until
	     * we encounter superclass or run out of nodes to test.
	     */

	    boolean DEBUG = false;
	    if (DEBUG) System.out.println(this+".subClassOf("+superclass+")");
	    
	    if (DEBUG) this.printInfo();

	    if (DEBUG) {
		for (Iterator i = superclass.iterator(); i.hasNext(); ) {
		    Object o = i.next();
		    System.out.println(o.getClass()+": "+o);
		}
	    }

	    // If there are no superclasses then return false.
	    if (superclass.size() == 0) {
		return(false);
	    }

	    // Keep track of all nodes already tested (to prevent an infinite loop)
	    Set tested = new HashSet();
	
	    // create set with initial Node and its equivalents
	    Set nodes = new HashSet(this.equivalent);
	    nodes.add(this);

	    // keep testing classes in nodes until we've reached the top of the class hierarchy (nodes is empty)
	    while (!nodes.isEmpty()) {

		if (DEBUG) System.out.println("Equivalent Nodes:");
		for (Iterator i = nodes.iterator(); i.hasNext();){
		    Node n = (Node)i.next();
		    if (DEBUG) System.out.println(n.getClass()+": "+n);
		}

		// Test to see if any is equal to the class node
		for (Iterator i = nodes.iterator(); i.hasNext(); ) {
		    if (superclass.contains((Node)i.next())) {
			return(true);
		    }
		}

		// Keep track of the nodes we've already checked
		tested.addAll(nodes);

		// Get all the parent nodes of the nodes we just checked.
		nodes = parentNodes(nodes);
	
		if (DEBUG) {
		    System.out.println("Parent Nodes:");
		    for (Iterator i = nodes.iterator(); i.hasNext(); System.out.println(i.next()));
		}
	    
		// Find all equivalent nodes
		nodes = equivalentNodes(nodes);

		if (DEBUG) {
		    System.out.println("Parent and Equivalent Nodes:");
		    for (Iterator i = nodes.iterator(); i.hasNext(); System.out.println(i.next()));
		}
	    
		// remove any nodes we've already checked
		nodes.removeAll(tested);

		if (DEBUG) {
		    System.out.println("Parent + Equivalent - Tested Nodes:");
		    for (Iterator i = nodes.iterator(); i.hasNext(); System.out.println(i.next()));
		}

	    }

	    // If we've not returned yet, then the node is not a subclass of the class node.
	    return(false);
	
	}

	/**
	 * Returns a set of restriction nodes for the class Node.
	 */
	// What about equivalence - should restrictions also be moved over?
	public Set getRestrictions(boolean useEquivalence) 
	    throws RDFException
	{
	    //System.out.println("DEBUG: In getRestrictions("+useEquivalence+") for node " + this);

	    // Create a set to store the restictions
	    Set ancestors = this.getAncestors(useEquivalence);
	    HashSet result = new HashSet();

	    // For each ancestor...
	    for (Iterator i = ancestors.iterator(); i.hasNext(); ) {
		Node n = (Node)i.next();
		// System.out.println("DEBUG: "+n+" type is "+n.getType());
		// If the node class is a restiction
		if ((n.getType() != null) && (restrictionClasses.contains(n.getType().resource))) {
		    // add to result set
		    result.add(n);
		}
	    }
	    //System.out.println("DEBUG: getRestrictions result: "+result);
	    return(result);
	}

	/*
	 * Returns all restrictions for a particular predicate.

	public Set getPredicateRestrictions(Node predicate, boolean useEquivalence) 
	throws RDFException
	{
	    System.out.println("DEBUG: in getPredicateRestrictions for Node "+this+" with predicate "+predicate);
            HashSet result = new HashSet();
	    Set restrictions = this.getRestrictions(useEquivalence);
	    
	    for (Iterator i = restrictions.iterator(); i.hasNext(); ) {
		Node r = (Node)i.next();
		if (r.getRestrictionProperty().equals(predicate)) {
		    result.add(r);
		}
	    }		    

	    System.out.println("DEBUG: getPredicateRestrictions results: "+result);
	    return(result);
	}
	*/


	/**
	 * Returns a set of ancestors for the current Node.
	 */
	// Take care of cycles
	// Take care of equivalence (as a flag)
	public Set getAncestors(boolean useEquivalence)
	    throws RDFException
	{
	    boolean DEBUG = false;
	    HashSet result = new HashSet();
	 
	    if (DEBUG) System.out.println("-- getAncestors("+useEquivalence+") for Node "+this);

	    Stack todo = new Stack();
	    todo.push(this);

	    // Pop from todo and put on results list
	    while (!todo.empty()) {
		Node n = (Node)todo.pop();
		result.add(n);
		// Add parent nodes to todo
		for (Iterator i = n.superclasses.iterator(); i.hasNext(); ) {
		    Node parent = (Node)i.next();
		    if (!result.contains(parent) && !todo.contains(parent)) {	
			if (DEBUG) System.out.println("adding superclass "+parent);
			todo.push(parent);
		    }
		}
		// Add equivalent nodes to todo
		if (useEquivalence) {
		    for (Iterator j = n.equivalent.iterator(); j.hasNext(); ) {
			Node equiv = (Node)j.next();
			if (!result.contains(equiv) && !todo.contains(equiv)) {
			    if (DEBUG) System.out.println("adding equivalent node "+equiv);
			    todo.push(equiv);
			}
		    }
		}
	    }

	    // Remove the original node and any equivalent nodes
	    result.remove(this);
	    if (useEquivalence) {
		for (Iterator i = equivalent.iterator(); i.hasNext(); result.remove(i.next())) {}
	    }
		    
	    if (DEBUG) System.out.println("-- getAncestors returning "+result);
	    return(result);
	}

	/**
	 * Validates an instance of this class against the restrictions of the class.
	 */
	/*
	public void validateInstance(Node instance) 
	throws RDFException
	{
	    if (!instance.getType().subClassOf(this)) {
		System.out.println("DEBUG: in validateInstance: Node "+instance+" is not of type "+this);
	    } else {
		// get all restrictions for this class
		Set restriction = this.getRestrictions(false);
		// For each restriction, validate the instance
		for (Iterator i = restriction.iterator(); i.hasNext();) {
		    ((Node)i.next()).validateInstance(instance);
		}
	    }
	}
	*/

	/**
	 * Returns the property that a restriction is on.
	 */
	public Node getProperty()
	    throws RDFException 
	{
	    Object prop = null;
	    try {
		prop = this.findAny(">onProperty");
	    } catch (Exception ex) {
		// ignore error
	    }

	    if (prop == null) {
		return(null);
	    } else if (prop instanceof DAMLModelImpl.Node) {
		return((DAMLModelImpl.Node)prop);
	    } else {
		return(null);
	    }
	}

	public Collection getDomainStatements()
	    throws RDFException
	{
	    //return getAllStatements(this.outgoing, domainPredicates);
	    return getOutgoingStatements(domainPredicates, true, true);
	}

		    
	public Vector getDomain() 
	    throws RDFException
	{
	    boolean DEBUG = false;
	    if (DEBUG) System.out.println("-----getDomain()");
	    //Vector domains = getAllStatements(this.outgoing, domainPredicates);
	    Vector domains = getOutgoingStatements(domainPredicates, true, true);
	    Vector result = new Vector();
	    for (Enumeration e = domains.elements(); e.hasMoreElements();) {
		Statement statement = (Statement) e.nextElement();
		if (DEBUG) System.out.println("-gD:Domain: "+statement);
		Object object = statement.getObject();
		if (object instanceof Resource) {
		    // Classes are valid domains
		    Node node = getNode((Resource)object, false);
		    if (node.isClassNode()) {
			if (DEBUG) System.out.println("-gD:Found "+node);
			result.add(node);
		    }
		}
	    }
	    return(result);
	}


	/**
	 * Returns a vector of ClassNodes for the Range of this PropertyNode (predicate)
	 */
	public Vector getRange()
	    throws RDFException
	{
	    boolean DEBUG = false;
	    if (DEBUG) System.out.println("-----getRange()");
	    //Vector ranges = getAllStatements(this.outgoing, rangePredicates);
	    Vector ranges = getOutgoingStatements(rangePredicates, true, true);
	    Vector result = new Vector();
	    for (Enumeration e = ranges.elements(); e.hasMoreElements();) {
		Statement statement = (Statement) e.nextElement();
		if (DEBUG) System.out.println("-gR:Range: "+statement);
		Object object = statement.getObject();
		if (DEBUG) System.out.println("-gR:Range Object type: "+object.getClass());
		if (object instanceof Resource) {
		    Node node = getNode((Resource)object, false);
		    if (node.isClassNode() || node.isDatatypeNode()) {
			if (DEBUG) System.out.println("-gR:Found "+node);
			result.add(node);
		    }
		}
		    
	    }
	    return(result);
	}

	/**
	 * Returns true if the Node is a class node.
	 */
	public boolean isClassNode()
	{
	    boolean DEBUG = false;
	    if (DEBUG) System.out.println("-----isClassNode");
	    // A Node is a class node if it's type is class
	    Set types = getTypes();
	    boolean isClass = false;
	    for (Iterator i = types.iterator(); (i.hasNext() && !isClass); ) {
		Node type = (DAMLModelImpl.Node)i.next();
		if (DEBUG) System.out.println("-iCN: "+type);
		if (classClasses.contains(type.resource)) {
		    isClass = true;
		}
	    }
	    if (DEBUG) System.out.println("-iCN: "+isClass);
	    return(isClass);
	}

	/**
	 * Returns true if this Node is a XML Schema Simple Datatype.
	 */
	public boolean isDatatypeNode ()
	    throws RDFException 
	{
	    return datatypeValidator.isAKnownSimpleType(resource.getNameSpace(), resource.getLocalName());
	}

	/**
	 * Returns true of this Node is an instance of one of the superclasses.
	 * @param superclasses a Vector of Nodes
	 */
	private boolean isInstanceOf (Vector superclasses) 
	{
	    boolean DEBUG = false;
	    if (DEBUG) System.out.println("== "+this+".isInstanceOf("+superclasses+")");
	    // Get types of this node
	    Set types = this.getTypes();
	    if (DEBUG) System.out.println("  Instance Types: "+types);
	    // One of the types must be a subclass of the class.
	    boolean isInstance = false;
	    for (Iterator i = types.iterator(); (i.hasNext() && !isInstance); ) {
		Node type = (Node)i.next();
		if (type.subClassOf(superclasses)) {
		    isInstance = true;
		}		
	    } 
	    if (DEBUG) System.out.println("== isInstanceOf returned "+isInstance);
	    return(isInstance);
	}

	/**
	 * Returns true of this Node is an instance of one of the superclasses.
	 * @param superclasses a Vector of Resources
	 */
	public boolean isInstanceOfResources (Vector superclasses) 
	{
	    boolean DEBUG = false;
	    if (superclasses.equals(resourceClasses)) {
		// Special Case for RDFS:Resource
		return true;
	    }
	    Vector nodes = new Vector();
	    for (Iterator i = superclasses.iterator(); i.hasNext(); ) {
		Resource r = (Resource)i.next();
		Node n = getNode(r);
		if (n != null) {
		    nodes.add(n);
		} else {
		    if (DEBUG) System.out.println("isInstanceOfResources: resource has no node ("+r+")");
		}
	    }
	    if (DEBUG) {
		System.out.println("isInstanceOfResources("+superclasses+")");
		System.out.println("checking "+nodes);
	    }
	    return this.isInstanceOf(nodes);
	}

	/**
	 * Returns true if this Node is an instance of the class
	 */
	public boolean isInstanceOf (Node superclass)
	{
	    if (resourceClasses.contains(superclass.resource)) {
		// Special Case for RDFS:Resource
		return true;
	    }
	    Vector v = new Vector();
	    v.add(superclass);
	    return(isInstanceOf(v));
	}

	public Statement getTypeStatement()
	{
	    return(findAnyStatement(outgoing, typePredicates));
	}
	
	/**
	 * Returns true if the node has a type that is a Datatype.
	 * For detecting typed literals.
	 */
	public boolean isDatatypeInstance() 
	    throws RDFException
	{
	    boolean DEBUG = false;
	    if (DEBUG) System.out.println("== isDatatypeInstance()");
	    boolean result = false;
	    Set types = getTypes();
	    for (Iterator i = types.iterator(); i.hasNext(); ) {
		Node type = (Node)i.next();
		String namespace = type.resource.getNameSpace();
		if (datatypeValidator.hasNamespace(namespace)) {
		    result = true;
		}
	    }
	    if (DEBUG) System.out.println("== isDatatypeInstance returned "+result);
	    return result;
	}

	
	/**
	 * Returns the first type Node of the instance that is a datatype.
	 */
	public Node getDatatypeInstanceType() 
	    throws RDFException
	{
	    boolean DEBUG = false;
	    if (DEBUG) System.out.println("== getDatatypeInstanceType()");
	    Set types = getTypes();
	    Node result = null;
	    for (Iterator i = types.iterator(); i.hasNext() && (result == null); ) {
		Node type = (Node)i.next();
		String namespace = type.resource.getNameSpace();
		if (datatypeValidator.hasNamespace(namespace)) {
		    result = type;
		}
	    }
	    if (DEBUG) System.out.println("== getDatatypeInstanceNode returned "+result);
	    return result;
	}

	/**
	 * Returns a set of RDFNodes from the resource via the rdf:value predicate.
	 */
	public Set getValues()
	{
	    HashSet result = new HashSet();
	    //Vector stmts = getAllStatements(this.outgoing, valuePredicates);
	    Vector stmts = getOutgoingStatements(valuePredicates, true, true);
	    for (Enumeration e = stmts.elements(); e.hasMoreElements(); ) {
		Statement statement = (Statement)e.nextElement();
		Object object = statement.getObject();
		result.add(object);
	    }
	    return result;
	}

	public Collection getValueStatements()
	{
	    //return getAllStatements(this.outgoing, valuePredicates);
	    return getOutgoingStatements(valuePredicates, true, true);
	}	

	public Collection getSubPropertyOfStatements()
	{
	    //return getAllStatements(this.outgoing, subPropertyOfPredicates);
	    return getOutgoingStatements(subPropertyOfPredicates, true, true);
	}

	public Collection getRangeStatements()
	{
	    //return getAllStatements(this.outgoing, rangePredicates);
	    return getOutgoingStatements(rangePredicates, true, true);
	    //return getOutgoingStatements(rangePredicates, useInheritance, useEquivalence);
	}

	public Collection getSamePropertyAsStatements()
	{
	    //return getAllStatements(this.outgoing, samePropertyAsPredicates);
	    return getOutgoingStatements(samePropertyAsPredicates, true, true);
	}

	public Collection getEquivalentToStatements()
	{
	    //return getAllStatements(this.outgoing, equivalentToPredicates);
	    return getOutgoingStatements(equivalentToPredicates, true, true);
	}

	public boolean isType(Collection typeCollection) 
	    throws RDFException 
	{
	    boolean DEBUG = false;
	    if (DEBUG) System.out.println("--"+this+".isType("+typeCollection+")");
	    Set types = getTypes();
	    if (DEBUG) System.out.println("  node has types "+types);
	    for (Iterator i = types.iterator(); i.hasNext(); ) {
		Resource r = ((Node)i.next()).resource;
		if (typeCollection.contains(r)) {
		    return true;
		}
	    }
	    return false;
	}

	/**
	 * Determine outgoing statements for the Node.  
	 * Use of inheritance and equivalence can be turned on or off.
	 */
	public Vector getOutgoingStatements(Collection predicateResources, boolean useInheritance, boolean useEquivalence) 
	{
	    boolean DEBUG = false;
	    Vector result = new Vector();

	    if (DEBUG) System.out.println("-- "+this+".getOutgoingStatements("+predicateResources+", "+
					  useInheritance+", "+useEquivalence+")");

	    try {
		// Determine the set of nodes to inherit from.
		Set nodes = null;
		if (useInheritance) {
		    nodes = getAncestors(useEquivalence);
		    nodes.add(this);
		    if (useEquivalence) {
			nodes.addAll(this.equivalent);
		    }
		} else if (useEquivalence) {
		    nodes = this.equivalent;
		} else {
		    nodes = new HashSet();
		    nodes.add(this);
		}

		if (DEBUG) System.out.println("nodes = "+nodes);

		// For each node, get the outgoing statements for the set of predicates
		for (Iterator i = nodes.iterator(); i.hasNext(); ) {
		    Node n = (Node)i.next();
		    result.addAll(getAllStatements(n.outgoing, predicateResources));
		}
	    } catch (RDFException e) {
		System.err.println("Ignoring error in getOutgoingExceptions:\n"+e);
	    }

	    if (DEBUG) System.out.println("-- getOutgoingStatements returning "+result);
	    
	    return result;
	}

	public Vector getOutgoingStatements(Resource predicateResource, boolean useInheritance, boolean useEquivalence) 
	{
	    HashSet predicates = new HashSet();
	    predicates.add(predicateResource);

	    return getOutgoingStatements(predicates, useInheritance, useEquivalence);
	}

	public Vector getOutgoingStatements(Node predicateNode, boolean useInheritance, boolean useEquivalence) 
	{
	    HashSet predicates = new HashSet();
	    predicates.add(predicateNode.resource);

	    return getOutgoingStatements(predicates, useInheritance, useEquivalence);
	}



    }
    // End of Node class definition
    
    public class Restriction 
    {
	public Node node;
	public Node onProperty;
	public Node toClass;
	public Object hasValue;
	public Node hasClass;
	public int minCardinality = -1;
	public int maxCardinality = -1;
	public int minCardinalityQ = -1;
	public int maxCardinalityQ = -1;
	public Node hasClassQ;

	/**
	 * Restrictions are created from Nodes. 
	 * Assumes that the Node has already been validated.
	 * Do we need to handle multiple values for toClass, hasValue, hasClass, and hasClassQ?
	 */
	Restriction (Node node)
	{
	    this.node = node;

	    Statement tmp = findAnyStatement(node.outgoing, onPropertyPredicates);
	    if ((tmp != null) && (tmp.getObject() instanceof Resource)) {
		onProperty = getNode((Resource)tmp.getObject());
	    }

	    tmp = findAnyStatement(node.outgoing, toClassPredicates);
	    if ((tmp != null) && (tmp.getObject() instanceof Resource)) {
		toClass = getNode((Resource)tmp.getObject());
	    }

	    tmp = findAnyStatement(node.outgoing, hasValuePredicates);
	    // Can be a Node or a datatype
	    if ((tmp != null)) {
		if (tmp.getObject() instanceof Resource) {
		    hasValue = getNode((Resource)tmp.getObject());
		} else {
		    hasValue = tmp.getObject();
		}
	    }

	    tmp = findAnyStatement(node.outgoing, hasClassPredicates);
	    if ((tmp != null) && (tmp.getObject() instanceof Resource)) {
		hasClass = getNode((Resource)tmp.getObject());
	    }


	    tmp = findAnyStatement(node.outgoing, cardinalityPredicates);
	    if ((tmp != null) && (tmp.getObject() instanceof Literal)) {
		try {
		    int card = Integer.parseInt(((Literal)tmp.getObject()).toString());
		    maxCardinality = card;
		    minCardinality = card;
		} catch (NumberFormatException e) {
		    // ignore error, checked elsewhere
		}
	    }

	    tmp = findAnyStatement(node.outgoing, minCardinalityPredicates);
	    if ((tmp != null) && (tmp.getObject() instanceof Literal)) {
		try {
		    int tmpCard = Integer.parseInt(((Literal)tmp.getObject()).toString());
		    minCardinality = tmpCard;
		} catch (NumberFormatException e) {
		    // ignore error, checked elsewhere
		}
	    }
	    tmp = findAnyStatement(node.outgoing, maxCardinalityPredicates);
	    if ((tmp != null) && (tmp.getObject() instanceof Literal)) {
		try {
		    int tmpCard = Integer.parseInt(((Literal)tmp.getObject()).toString());
		    maxCardinality = tmpCard;
		} catch (NumberFormatException e) {
		    // ignore error, checked elsewhere
		}
	    }
		
	    tmp = findAnyStatement(node.outgoing, hasClassQPredicates);
	    if ((tmp != null) && (tmp.getObject() instanceof Resource)) {
		hasClassQ = getNode((Resource)tmp.getObject());
	    }

	    tmp = findAnyStatement(node.outgoing, cardinalityQPredicates);
	    if ((tmp != null) && (tmp.getObject() instanceof Literal)) {
		try {
		    int card = Integer.parseInt(((Literal)tmp.getObject()).toString());
		    maxCardinalityQ = card;
		    minCardinalityQ = card;
		} catch (NumberFormatException e) {
		    // ignore error, checked elsewhere
		}
	    }
	    tmp = findAnyStatement(node.outgoing, minCardinalityQPredicates);
	    if ((tmp != null) && (tmp.getObject() instanceof Literal)) {
		try {
		    int card = Integer.parseInt(((Literal)tmp.getObject()).toString());
		    minCardinalityQ = card;
		} catch (NumberFormatException e) {
		    // ignore error, checked elsewhere
		}
	    }
	    tmp = findAnyStatement(node.outgoing, maxCardinalityQPredicates);
	    if ((tmp != null) && (tmp.getObject() instanceof Literal)) {
		try {
		    int card = Integer.parseInt(((Literal)tmp.getObject()).toString());
		    maxCardinalityQ = card;
		} catch (NumberFormatException e) {
		    // ignore error, checked elsewhere
		}
	    }
	}
	
	public String toString()
	{
	    String s = "Restriction:\n";
	    s += "node "+node+"\n";
	    s += "onProperty "+onProperty+"\n";
	    s += "toClass "+toClass+"\n";
	    s += "hasValue "+hasValue+"\n";
	    s += "hasClass "+hasClass+"\n";
	    s += "minCardinality "+minCardinality+"\n";
	    s += "maxCardinality "+maxCardinality+"\n";
	    s += "minCardinalityQ "+minCardinalityQ+"\n";
	    s += "maxCardinalityQ "+maxCardinalityQ+"\n";
	    s += "hasClassQ "+hasClassQ+"\n";

	    return(s);
	}

    } // End of Restriction class definition

    /**
     * Map from Resource to Node.
     */
    java.util.Hashtable nodes = new java.util.Hashtable();
	
    /**
     * Finds the Node for the given resource, possibly creating a new Node.
     * @param resource Resource to find.
     * @param create If true, a new Node will be created if none is found.
     * @returns The Node for the Resource, or null if not found an create is false.
     */
    public Node getNode(Resource resource,
			boolean create)
    {
	Node retval = (Node) nodes.get(resource);
	if ((retval == null) && create)
	    {
		retval = new Node(resource);
		nodes.put(resource, retval);
		//System.out.println(this+" CREATED NODE "+retval);
	    }
	return retval;
    }

    /**
     * Find the node for the given Resource.
     * @param resource Resource to find.
     * @return Returns the Node for the Resource or null if no Node exists.
     */
    public Node getNode(Resource resource)
    {
	return getNode(resource, false);
    }

    /**
     * Merge the equivalence relations between the two nodes (and all equivalent nodes).
     * Should this be a Node function? 
     */
    protected void mergeEquivalenceSets(Node n1, Node n2) {
	Set set1 = new HashSet(n1.equivalent);
	set1.add(n1);
	Set set2 = new HashSet(n2.equivalent);
	set2.add(n2);
	
	for (Iterator i = set1.iterator(); i.hasNext(); ) {
	    Node n = (Node)i.next();
	    n.equivalent.addAll(set2);
	}

	for (Iterator i = set2.iterator(); i.hasNext(); ) {
	    Node n = (Node)i.next();
	    n.equivalent.addAll(set1);
	}
    }


    /**
     * Create a new Node centered view from a Model.
     */
    public DAMLModelImpl(Model model, DAMLDatatypeValidator datatypeValidator,
			 Preferences preferences)
	throws RDFException
    {
	this.model = model;
	this.datatypeValidator = datatypeValidator;
	this.preferences = preferences;
	try {
	    setDAMLVocabulary(DAML_VOCABULARY);
	} catch (java.lang.ClassNotFoundException e) {
	    System.err.println("WARNING: "+DAML_VOCABULARY+" class not found.\nCheck classpath or use setDAMLVocabulary().");
	}

	createNodesFromModel(model);
    }

    public DAMLModelImpl(DAMLDatatypeValidator datatypeValidator, Preferences preferences)
    {
	this.datatypeValidator = datatypeValidator;
	this.preferences = preferences;
	try {
	    setDAMLVocabulary(DAML_VOCABULARY);
	} catch (java.lang.ClassNotFoundException e) {
	    System.err.println("WARNING: "+DAML_VOCABULARY+" class not found.\nCheck classpath or use setDAMLVocabulary().");
	}
    }

    public void setDAMLVocabulary(String DAMLVocabulary) 
	throws java.lang.ClassNotFoundException
    {
	boolean DEBUG = false;
	if (DEBUG == true) System.out.println("Setting DAMLVocabulary to "+DAMLVocabulary);
	Class vocab = java.lang.Class.forName(DAMLVocabulary);
	this.DAMLVocabulary = vocab;
    }


    protected void createNodesFromModel (Model model) 
	throws RDFException
    {
	for (StmtIterator en = model.listStatements(); en.hasNext();)
	    {
		Statement statement = (Statement) en.next();
		Resource subject = statement.getSubject();
		RDFNode object = statement.getObject();
		
		Node subjectNode = getNode(subject, true);
		subjectNode.outgoing.add(statement);

		Node objectNode = null;
		if (object instanceof Resource)
		    {
			objectNode = getNode((Resource) object, true);
			objectNode.incoming.add(statement);
		    }

		// Add equivalence Nodes
		Resource predicate = statement.getPredicate();
		if ((objectNode != null) && (equivalentPredicates.contains(predicate))) {
		    // Merge the equivalence sets of both Nodes (and their equivalent Nodes)
		    mergeEquivalenceSets(subjectNode, objectNode);
		}
		

		// Add subclass Nodes
		if ((objectNode != null) && (ancestorPredicates.contains(predicate))) {
		    // add parent/child relationships between nodes 
		    // System.out.println(">>> "+subjectNode+" is a subclass of "+objectNode);
		    subjectNode.superclasses.add(objectNode);
		    objectNode.subclasses.add(subjectNode);
		}
	    }
    }

    /**
     * Add the RDF Model to the DAML Model.
     */
    public void addModel (Model rdfModel) 
	throws RDFException
    {
	// If there is no current model, make a new one.
	if (model == null) {
	    model = new ModelMem();
	}
	// Add each statement separately because we don't want to change the original Model
	for (StmtIterator en = rdfModel.listStatements(); en.hasNext(); ) {
	    // Add the statement to the local Model
	    model.add((Statement)en.next());
	}

	// Create Nodes (or update existing Nodes) from the statements in the model.
	createNodesFromModel(rdfModel);
    }


    /**
     * Finds all instances of the specified class.
     * Note: Ignores the namespace portion of the class URI.
     * @return a Vector of Nodes that are instances of the class.
     */
    public java.util.Vector allInstances(String cl)
	throws Exception
    {
	java.util.Vector retval = new java.util.Vector();

	cl = "#" + cl;

	for (StmtIterator en = model.listStatements(); en.hasNext();)
	    {
		Statement statement = (Statement) en.next();
		if (statement.getPredicate().equals(com.hp.hpl.mesa.rdf.jena.vocabulary.RDF.type))
		    {
			if (statement.getObject().toString().endsWith(cl))
			    retval.add(nodes.get(statement.getSubject()));
		    }
	    }

	return retval;
    }

    /**
     * Return all the nodes in the DAML Model
     */

    public Enumeration getNodes()
    {
	return(nodes.elements());
    }

    public static Vector getAllStatements(Model model, Resource predicate) {
	Vector result = new Vector();
	try {
	    for (StmtIterator e = model.listStatements(); e.hasNext();) {
		Statement statement = (Statement) e.next();
		if (predicate.equals(statement.getPredicate())) {
		    result.add(statement);
		}
	    } 
	} catch (RDFException me) {
	    System.out.println("DEBUG: Model Exception occurred in getAllStatements (Ignoring):\n"+me);
	}
	return(result);
    }	

    public Vector getAllStatements(Model model, Node predicate) {
	return(getAllStatements(model, predicate.resource));
    }

    /**
     * Returns a vector of from the model that match the collection of predicate resources.
     */
    public static Vector getAllStatements(Model model, Collection predicates) {
	Vector result = new Vector();
	try {
	    for (StmtIterator e = model.listStatements(); e.hasNext();) {
		Statement statement = (Statement) e.next();
		if (predicates.contains(statement.getPredicate())) {
		    result.add(statement);
		}
	    } 
	} catch (RDFException me) {
	    System.out.println("DEBUG: Model Exception occurred in getAllStatements (Ignoring):\n"+me);
	}
	return(result);
    }
		

    /**
	 * Given a vector of nodes, returns a vector of all equivantent nodes.  
	 * The resulting list includes the nodes from the initial list.
	 */
    public static Set equivalentNodes(Set nodes)
    {
	Set equiv = new HashSet();
	for (Iterator i = nodes.iterator(); i.hasNext(); ) {
	    Node node = (Node)i.next();
	    equiv.addAll(node.equivalent);
	    equiv.add(node);
	}
	return(equiv);
    }


    /**
     * Given a vector of nodes, returns a vector of all parent classes for those nodes.
     */
    public static Set parentNodes(Set nodes) 
    {
	Set superclasses = new HashSet();
	for (Iterator i = nodes.iterator(); i.hasNext(); ) {
	    superclasses.addAll(((Node)i.next()).superclasses);
	}
	return(superclasses);
    }

    public Restriction getRestriction(Node node) 
    {
	return(new Restriction(node));
    }


	/**
	 * Returns the first statement from the model that uses the given predicate.
	 * Note: this ignores the URI.
	 */
	public Statement findAnyStatement(Model model,
							     String predicate)
	{
	    String match = "#" + predicate;

	    try {
		for (StmtIterator en = model.listStatements(); en.hasNext();)
		    {
			Statement statement = (Statement) en.next();
			String thisPredicate = statement.getPredicate().toString();
			if (thisPredicate.endsWith(match))
			    return statement;
		    }
	    } catch (RDFException me) {
		//Some error processing the model
		System.out.println("RDFException occurred in findAnyStatement (Ignoring):\n"+me.getLocalizedMessage());
	    }
	    return null;
	}

	/**
	 * This version matches the complete URI of the predicate.
	 */
	public Statement findAnyStatement(Model model,
							     Resource predicate)
	{
	    try {
		for (StmtIterator en = model.listStatements(); en.hasNext();)
		    {
			Statement statement = (Statement) en.next();
			Resource thisPredicate = statement.getPredicate();
			if (thisPredicate.equals(predicate))
			    return statement;
		    }
	    } catch (RDFException me) {
		// Some error processing the model
		System.out.println("RDFException occurred in findAnyStatement (Ignoring):\n"+me.getLocalizedMessage());
	    }
	    return null;
	}

	/**
	 * This version matches the complete URI of any of a set of predicates.
	 */
	public Statement findAnyStatement(Model model,
							       Vector predicates)
	{
	    try {
		for (StmtIterator en = model.listStatements(); en.hasNext();)
		    {
			Statement statement = (Statement) en.next();
			Resource thisPredicate = statement.getPredicate();
			if (predicates.contains(thisPredicate))
			    return statement;
		    }
	    } catch (RDFException me) {
		// Some error processing the model
		System.out.println("RDFException occurred in findAnyStatement (Ignoring):\n"+me.getLocalizedMessage());
	    }
	    return null;
	}

    public boolean isInstanceOf(RDFNode instance, Node superclass) 
	throws RDFException
    {
	boolean DEBUG = false;
	boolean ERROR = true;
	if (DEBUG) System.out.println("== isInstanceOf("+instance+", "+superclass+")");
	String namespace = superclass.resource.getNameSpace();
	if (datatypeValidator.hasNamespace(namespace)) {
	    // the superclass is a datatype
	    if (DEBUG) System.out.println("   superclass is a datatype");
	    String datatypeName = superclass.resource.getLocalName();
	    boolean result = true;
	    if (instance instanceof Literal) {
		if (DEBUG) System.out.println("   instance is a Literal");
		try {
		    result = datatypeValidator.isAValidValue(instance.toString(), namespace, datatypeName);
		} catch (UnknownSimpleTypeException e3) {
		    // Type doesn't exist, can't do test.  return true
		    result = true;
		}
	    } else if (instance instanceof Resource) {
		if (DEBUG) System.out.println("    instance is a Resource");
		Node inst = getNode((Resource)instance, false);
		if (inst.isDatatypeInstance()) {
		    if (DEBUG) System.out.println("   instance is a datatype");
		    // Get the datatype info
		    Node valType = inst.getDatatypeInstanceType();
		    String valNamespace = valType.resource.getNameSpace();
		    String valDatatype = valType.resource.getLocalName();
		    // Get value(s?)
		    Set values = inst.getValues();
		    int size = values.size();
		    if (DEBUG) System.out.println("   instance has "+size+" value(s)");
		    if (size == 1) {
			RDFNode val = (RDFNode)(values.iterator().next());
			if (val instanceof Literal) {
			    if (DEBUG) System.out.println("   instance has value "+((Literal)val).toString());
			    try {
				result = datatypeValidator.isAValidTypedValue(valNamespace, valDatatype, ((Literal)val).toString(), namespace, datatypeName);
			    } catch (UnknownSimpleTypeException e) {
				// Type doesn't exist, can't do test.  Return true
				if (DEBUG) System.out.println("    UnknownSimpleTypeException "+e);
				result = true;
			    }
			} else {
			    if (DEBUG) System.out.println("   instance is not a Literal");
			    result = false;
			}
		    } else if (size == 0) {
			if (ERROR) System.out.println("isInstanceOf("+instance+", "+superclass+") - Instance has no value.");
		    } else {
			if (ERROR) System.out.println("isInstanceOf("+instance+", "+superclass+") - Instance has multiple values.");
		    }
		} else {
		    // Instance is a Resource but not a Typed Literal
		    if (DEBUG) System.out.println("   instance is a Resource but not a Typed Literal");
		    result = false;
		}
	    } else {
		// Not a Literal or Resource - what is it?
		if (ERROR) System.out.println("isInstanceOf("+instance+", "+superclass+") - Unexpected instance, not a Literal or Resource.");
		return false;
	    }
	    if (DEBUG) System.out.println("== isInstanceOf returned "+result);
	    return result;
	} else {
	    if (DEBUG) System.out.println("   superclass is NOT a datatype");
	    // Object Restriction
	    if (instance instanceof Literal) {
		if (literalClasses.contains(superclass.resource)) {
		    if (DEBUG) System.out.println("== isInstanceOf returned true because of literals");
		    return true;
		} else {
		    if (DEBUG) System.out.println("== isInstanceOf returned false because instance is a Literal");
		    return false;
		}
	    } else if (instance instanceof Resource) {
		if (DEBUG) System.out.println("   instance is Resource");
		if (resourceClasses.contains(superclass.resource)) {
		    // Special Case for RDFS:Resource
		    return true;
		}
		Node instanceNode = getNode((Resource)instance);
		if (instanceNode == null) {
		    System.out.println("DEBUG: "+instance+" doesn't have a node.");
		    return (false);
		} else {
		    if (DEBUG) System.out.println("== isInstanceOf is calling "+instanceNode+".isIntanceOf("+superclass+")");
		    return(instanceNode.isInstanceOf(superclass));
		}
	    } else {
		// Not a literal or resource
		System.err.println("WARNING: Unexpected value in isInstanceOf() : "+instance);
		return (false);
	    }
	}
    }
	

    /**
     * Generate a generic definition for a property.
     * Undefined propertys are assumed to be a binary property with no globally 
     * applicable constraints.
     */
    public void defineGenericProperty(Resource property)
	throws RDFException
    {
	// Create a Model to hold the statements we want to add.
	Model tmpModel = new ModelMem();
	
	// Get the node factory of the model
	//NodeFactory f = tmpModel.getNodeFactory();
	
	try {
	    // Create statement that defines the property
	    Statement s = tmpModel.createStatement(property, (Property)DAMLVocabulary.getField("type").get(DAMLVocabulary),
					    (Resource)DAMLVocabulary.getField("Property").get(DAMLVocabulary));

	    // Add statement to tmpModel
	    tmpModel.add(s);

	    // Catch and signal error caused by missing DAML resource objects
	} catch (NoSuchFieldException e2) {
	    System.err.println("Error in defineGenericProperty:\n"+e2);
	} catch (IllegalAccessException e3) {
	    System.err.println("Error in defineGenericProperty:\n"+e3);
	}
	
	// Add the statements in tmpModel to the current DAML model
	addModel(tmpModel);
    }


    public void addDatatypes(Collection datatypes)
	throws RDFException
    {
	// Create a Model to hold the statements we want to add.
	Model tmpModel = new ModelMem();

	// Get the Node Factory for the model
	//NodeFactory f = tmpModel.getNodeFactory();
	try {
	    // For each datatype
	    for (Iterator i = datatypes.iterator(); i.hasNext(); ) {
		String datatype = (String)i.next();
		// Create statements that say that the datatype is a Class and is a subclass of Datatype
		Resource dtResource = tmpModel.createResource(datatype);
		Statement s1 = tmpModel.createStatement(dtResource, (Property)DAMLVocabulary.getField("type").get(DAMLVocabulary),
						 (Resource)DAMLVocabulary.getField("Class").get(DAMLVocabulary));
		Statement s2 = tmpModel.createStatement(dtResource, (Property)DAMLVocabulary.getField("subClassOf").get(DAMLVocabulary), 
						 (Resource)DAMLVocabulary.getField("Datatype").get(DAMLVocabulary));
		// Add to the temporary model
		tmpModel.add(s1);
		tmpModel.add(s2);
	    }
	} catch (NoSuchFieldException e2) {
	    System.err.println("Error in addDatatypes:\n"+e2);
	} catch (IllegalAccessException e3) {
	    System.err.println("Error in addDatatypes:\n"+e3);
	}

	// Add the statements to the current DAML Model.
	addModel(tmpModel);
    }
}

